unit CDDatabase;

interface

uses
	ADODB, CDDBConnect, Classes, SysUtils;

type
	TADOConnectionEx = class(TADOConnection)
  	protected
    	function CreateQuery : TADOQuery; overload;
      function CreateQuery (ASQL : string): TADOQuery; overload;
      function CreateEmptyQuery(ATableName : string) : TADOQuery;
  end;

  TAccessDBEntry = class(TCDDBEntry)
  	protected
    	FDBConnection : TADOConnectionEx;
      procedure PopulateTracks;
    	function GetTracks: TStrings; override;
    public
    	property DBConnection : TADOConnectionEx read FDBConnection write
      	FDBConnection;
  end;

	TCDDatabase = class(TComponent)
  	private
    	FFileName: TFileName;
    	procedure SetFileName(const Value: TFileName);
  	protected
    	FDBConnection : TADOConnectionEx;
      FIDList : TStringList;
      procedure RefreshIDList;
    	procedure CreateDatabase(AFileName : string);
      procedure CreateTables;
      function GetConnectionString(AFileName : string) : string;
  	public
    	constructor Create(AOwner : TComponent); override;
      destructor Destroy; override;
      function RecordingCount : integer;
      procedure GetRecording(Index : integer; ARecording : TAccessDBEntry); overload;
      function GetRecording(ACDID : string; ARecording : TAccessDBEntry) : Boolean; overload;
      procedure AddRecording(ARecording : TCDDBEntry);
      procedure DeleteRecording(ACDID : string);
      procedure UpdateRecording(ARecording : TCDDBEntry);
      procedure SaveAsHTML(AFileName : TFileName);
      // properties
      property DBConnection : TADOConnectionEx read FDBConnection;
  end;

procedure Register;

////////////////////////////////////////////////////////////////////////////////

implementation

uses
	ADOX_TLB, HTTPApp, Forms, Controls, XMLDoc, XMLIntf;

const
	CRecCDID = 'CDID';
  CRecCDTitle = 'CDTitle';
  CRecArtistTitle = 'ArtistTitle';
  CTrackTrackTitle = 'TrackTitle';

procedure Register;
begin
  RegisterComponents('DoogalSoft', [TCDDatabase]);
end;

////////////////////////////////////////////////////////////////////////////////
{ TCDDatabase }

constructor TCDDatabase.Create(AOwner : TComponent);
begin
	inherited Create(AOwner);

  // open the connection
  FDBConnection := TADOConnectionEx.Create(nil);
  FDBConnection.LoginPrompt := False;

  FIDList := TStringList.Create;
  SetFileName(ExtractFilePath(Application.ExeName) + 'CDLibrary.mdb');
end;

////////////////////////////////////////////////////////////////////////////////

destructor TCDDatabase.Destroy;
begin
	FIDList.Free;
	FDBConnection.Free;
  inherited;
end;

////////////////////////////////////////////////////////////////////////////////
// creates the MDB file

procedure TCDDatabase.CreateDatabase(AFileName: string);
var
	LCatalog : _Catalog;
begin
  LCatalog := CoCatalog.Create;
  LCatalog.Create(GetConnectionString(AFileName));
  LCatalog := nil;
end;

////////////////////////////////////////////////////////////////////////////////

function TCDDatabase.GetConnectionString(AFileName: string): string;
begin
  Result := 'Provider=Microsoft.Jet.OLEDB.4.0;Data Source=' + AFileName;
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDDatabase.CreateTables;
begin
  FDBConnection.Execute('CREATE TABLE Recording (CDID TEXT(8), '+
  	'CDTitle TEXT(100), ArtistTitle TEXT(100), ' +
  	'CONSTRAINT PrimaryIndex PRIMARY KEY (CDID))');

  FDBConnection.Execute('CREATE TABLE Track (CDID TEXT(8), TrackNumber INT, '+
  	'TrackTitle TEXT(100), CONSTRAINT PrimaryIndex PRIMARY KEY (CDID, TrackNumber))');
end;

////////////////////////////////////////////////////////////////////////////////

function TCDDatabase.RecordingCount: integer;
var
	LCountQry : TADOQuery;
begin
  LCountQry := FDBConnection.CreateQuery('SELECT COUNT(CDID) FROM Recording');
  Result := LCountQry.Fields[0].AsInteger;
  LCountQry.Free;
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDDatabase.AddRecording(ARecording: TCDDBEntry);
var
	Loop : integer;
  LRecordingQry, LTrackQry : TADOQuery;
begin
	FDBConnection.BeginTrans;

  try
  	// insert recording info
  	LRecordingQry := FDBConnection.CreateEmptyQuery('Recording');
    LRecordingQry.Append;
    LRecordingQry.FieldByName(CRecCDID).AsString := ARecording.CDID;
    LRecordingQry.FieldByName(CRecCDTitle).AsString := ARecording.CDTitle;
    LRecordingQry.FieldByName(CRecArtistTitle).AsString := ARecording.ArtistTitle;
    LRecordingQry.Post;
    LRecordingQry.Free;

  	// insert tracks
    LTrackQry := FDBConnection.CreateEmptyQuery('Track');
		for Loop := 0 to ARecording.Tracks.Count-1 do
  	begin
      LTrackQry.Append;
      LTrackQry.FieldByName(CRecCDID).AsString := ARecording.CDID;
      LTrackQry.FieldByName('TrackNumber').AsInteger := Loop+1;
      LTrackQry.FieldByName(CTrackTrackTitle).AsString := ARecording.Tracks[Loop];
      LTrackQry.Post;
  	end;
    LTrackQry.Free;

    FDBConnection.CommitTrans;

    FIDList.Add(ARecording.CDID);
  except
    FDBConnection.RollbackTrans;
    raise;
  end;
end;

////////////////////////////////////////////////////////////////////////////////

function TCDDatabase.GetRecording(ACDID : string; ARecording: TAccessDBEntry): Boolean;
var
	LRecordingQuery : TADOQuery;
begin
	Result := False;
  LRecordingQuery := FDBConnection.CreateQuery(
  	'SELECT * FROM Recording WHERE CDID = ''' + ACDID + '''');
  try
  	if (LRecordingQuery.RecordCount = 1) then
  	begin
    	ARecording.FCDID := ACDID;
    	ARecording.FCDTitle := LRecordingQuery.FieldByName(CRecCDTitle).AsString;
    	ARecording.FArtistTitle := LRecordingQuery.FieldByName(CRecArtistTitle).AsString;
      ARecording.FDBConnection := FDBConnection;
      Result := True;
  	end;
  finally
  	LRecordingQuery.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDDatabase.DeleteRecording(ACDID: string);
begin
  FDBConnection.BeginTrans;
  try
  	FDBConnection.Execute('DELETE FROM Recording WHERE CDID = ''' +
  		ACDID + '''');
  	FDBConnection.Execute('DELETE FROM Track WHERE CDID = ''' +
  		ACDID + '''');
  	FDBConnection.CommitTrans;

    FIDList.Delete(FIDList.IndexOf(ACDID));
  except
    FDBConnection.RollbackTrans;
    raise;
  end;
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDDatabase.UpdateRecording(ARecording: TCDDBEntry);
begin
  DeleteRecording(ARecording.CDID);
  AddRecording(ARecording);
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDDatabase.SetFileName(const Value: TFileName);
var
	LNewDatabase : Boolean;
begin
	FDBConnection.Close;

	LNewDatabase := not FileExists(Value);
  if LNewDatabase then
  	CreateDatabase(Value);
  FFileName := Value;

  FDBConnection.ConnectionString := GetConnectionString(Value);
  FDBConnection.Open;

	if LNewDatabase then
  	CreateTables;

  RefreshIDList;
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDDatabase.GetRecording(Index: integer;
  ARecording: TAccessDBEntry);
begin
  GetRecording(FIDList[Index], ARecording);
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDDatabase.RefreshIDList;
var
	LQry : TADOQuery;
begin
	if FIDList.Count = 0 then
  begin
  	LQry := FDBConnection.CreateQuery('SELECT CDID, ArtistTitle, CDTitle FROM Recording ' +
    	'ORDER BY ArtistTitle, CDTitle');
    try
  		while not LQry.Eof do
    	begin
    		FIDList.Add(LQry.Fields[0].AsString);
  			LQry.Next;
  		end;
    finally
  		LQry.Free;
    end;
  end;
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDDatabase.SaveAsHTML(AFileName: TFileName);
const
	CCollectionTitle = 'CD Collection';
var
  LXMLDoc : TXMLDocument;
  LRecording: TAccessDBEntry;
  LHeader, LBody, LTable, LAttr, LTableRow : IXMLNode;
  Loop : integer;

begin
  Screen.Cursor := crHourglass;
  try
    LXMLDoc := TXMLDocument.Create(nil);
  	LRecording := TAccessDBEntry.Create;
  	try
      LXMLDoc.Active  := True;
      LXMLDoc.DocumentElement := LXMLDoc.CreateElement('html', '');

      // header
      LHeader := LXMLDoc.DocumentElement.AddChild('head');
      with LHeader.AddChild('title') do
        Text := CCollectionTitle;

      // body
      LBody := LXMLDoc.DocumentElement.AddChild('body');
      with LBody.AddChild('h2') do
        Text := CCollectionTitle;
      LBody.AddChild('hr');

      // start the table
      LTable := LBody.AddChild('table');
      LAttr := LXMLDoc.CreateNode('border', ntAttribute);
      LAttr.Text := '0';
      LTable.AttributeNodes.Add(LAttr);
      LAttr := LXMLDoc.CreateNode('width', ntAttribute);
      LAttr.Text := '100%';
      LTable.AttributeNodes.Add(LAttr);

      // each entry
      for Loop := 0 to RecordingCount-1 do
  		begin
        GetRecording(Loop, LRecording);
        LTableRow := LTable.AddChild('tr');
        with LTableRow.AddChild('td') do
          Text := LRecording.CDTitle;
        with LTableRow.AddChild('td') do
          Text := LRecording.ArtistTitle;
      end;

      LXMLDoc.SaveToFile(AFileName);
    finally
  		LRecording.Free;
  		LXMLDoc.Free;
  	end;
  finally
  	Screen.Cursor := crDefault;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
{ TAccessDBEntry }
////////////////////////////////////////////////////////////////////////////////

function TAccessDBEntry.GetTracks: TStrings;
begin
  Result := inherited GetTracks;
  PopulateTracks;
end;

////////////////////////////////////////////////////////////////////////////////

procedure TAccessDBEntry.PopulateTracks;
var
	LTrackQuery : TADOQuery;
begin
  if FTrackList.Count = 0 then
  begin
    // get track names
    LTrackQuery := FDBConnection.CreateQuery('SELECT * FROM Track WHERE CDID = ''' +
      FCDID + ''' ORDER BY TrackNumber');
    try
      while not LTrackQuery.Eof do
      begin
        FTrackList.Add(LTrackQuery.FieldByName(CTrackTrackTitle).AsString);
        LTrackQuery.Next;
      end;
    finally
      LTrackQuery.Free;
    end;
  end;
end;

////////////////////////////////////////////////////////////////////////////////

function TADOConnectionEx.CreateEmptyQuery(ATableName: string): TADOQuery;
begin
	Result := CreateQuery;
  Result.SQL.Add('SELECT * FROM ' + ATableName + ' WHERE 1=2');
	Result.Open;
end;

////////////////////////////////////////////////////////////////////////////////

function TADOConnectionEx.CreateQuery: TADOQuery;
begin
  Result := TADOQuery.Create(nil);
  Result.Connection := Self;
end;

////////////////////////////////////////////////////////////////////////////////

function TADOConnectionEx.CreateQuery (ASQL : string): TADOQuery;
begin
	Result := CreateQuery;
  Result.SQL.Text := ASQL;
  Result.Open;
end;

end.
