unit FormState;

{
 [FormState] [1.2]
Delphi 2005
February 2006


LICENSE

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
"http://www.mozilla.org/MPL/"

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is "[FormState.pas]".

The Initial Developer of the Original Code is Martin Holmes (Victoria,
BC, Canada, "http://www.mholmes.com/"). Copyright (C) 2005 Martin Holmes 
and the University of Victoria Computing and Media Centre. The code was 
co-developed for university and personal projects, and rights are shared
by Martin Holmes and the University of Victoria. All Rights Reserved.
}

{
 Class initially created by Martin Holmes Fall 2005, using Delpi 2005, and only tested with
 Delphi 2005.

 This enables the saving and loading of
 form position and size, and the size of a
 ny controls with an Align setting,
 and the settings of various other controls,
 in an XML or an INI file in the AppData folder.

 The object also publishes its AppDirPath property so any other code using
 it can use the same application data folder.

 Whether XML or INI format is used depends on the second parameter  of
 the Create function, the boolean UseXML.

 Use it like this:

 Create a private variable in the form:

      FFormStateSaver: TFormStateSaver;

 In the FormShow, do this:

      FFormStateSaver := TFormStateSaver.Create(Self, True);

 (This is the simplest constructor; the boolean  parameter specifies that
 XML should be used for storing the data, as opposed to INI file format.
 A more complex constructor is available, which enables you to set parameters
 for storing the states of a range of controls. See the code for details.)
 In the FormClose or the FormDestroy, do this:

      FreeAndNil(FFormStateSaver);

 This is specialized for TTntForms (from Troy Wolbrink's Unicode Controls)
 because that's what I use for all my forms in Delphi 2005.

 For this library to work properly:

  1. Form Position should be DefaultPosOnly. (However, if you are going to show the
  form modally, set it to poDesigned. If you don't, it will be positioned
  automatically by the system in the wrong place.)

  2. Application Title must be specified (in the project options).

  3. The application must include version info.

The latter two are required in order to save and load the file reliably.

Other properties allow the saving of other data.


 Dependencies:

 XDOM_3_1 (Dieter Kohler) for reading and writing XML files.

 VersionInfo (this has a class for getting version info
 about the running application). It would be feasible to use an instantiated
 TAppVersionInfo belonging to an instance of TSplashAbout owned by the main
 form, but then this library wouldn't be portable.

 GenFunctions (my library containing, among other things, code for getting
 Special Folder info from Windows).

 IniFiles (this could easily be replaced with RegIni or an XML equivalent).

 TntUnicode libraries (Troy Wolbrink).

}
interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ExtCtrls, StdCtrls, TntForms, TntStdCtrls, TntExtCtrls,
  VersionInfo, GenFunctions, IniFiles, XDOM_3_1, TntComCtrls, TntSysUtils,
  Spin;

type
  TFormStateSaver = class(TObject)

  private
    FOwnerForm: TTntForm;
    FVersionInfo: TAppVersionInfo;
    FProductID: WideString;
    FDirPath: WideString;
    FFilePath: WideString;
    FXMLFilePath: WideString;
    FControlList: TList;
    FUseXML: Boolean;
    FSaveCheckboxStates: Boolean;
    FSaveSpinEditValues: Boolean;
    FSaveComboBoxStates: Boolean;
    FSaveRadioGroupStates: Boolean;
    FSaveEditContents: Boolean;
    procedure ReadStateFromFile;
    procedure ReadStateFromINIFile;
    procedure ReadStateFromXMLFile;
    procedure WriteStateToFile;
    procedure WriteStateToINIFile;
    procedure WriteStateToXMLFile;
    procedure EnumerateControls(Parent: TComponent);
  public
    constructor Create(OwnerForm: TTntForm; UseXML: Boolean); overload;
    constructor Create(OwnerForm: TTntForm; UseXML: Boolean;
                      SaveCheckboxStates, SaveSpinEditValues,
                      SaveComboBoxStates, SaveRadioGroupStates,
                      SaveEditContents: Boolean); overload;
    destructor Destroy; override;
  published
    property AppDirPath: WideString read FDirPath;
    property UseXML: Boolean read FUseXML write FUseXML default True;
    property SaveCheckboxStates: Boolean read FSaveCheckboxStates write FSaveCheckboxStates default False;
    property SaveSpinEditValues: Boolean read FSaveSpinEditValues write FSaveSpinEditValues default False;
    property SaveComboBoxStates: Boolean read FSaveComboBoxStates write FSaveComboBoxStates default False;
    property SaveRadioGroupStates: Boolean read FSaveRadioGroupStates write FSaveRadioGroupStates default False;
    property SaveEditContents: Boolean read FSaveEditContents write FSaveEditContents default False;
  end;

implementation

{ TFormStateSaver }

constructor TFormStateSaver.Create(OwnerForm: TTntForm; UseXML: Boolean);
begin
  FOwnerForm := OwnerForm;
  FUseXML := UseXML;
//Other option properties default to false in this constructor.

//Build the application data path for storing the info
  FVersionInfo := TAppVersionInfo.Create;
  FProductID := FVersionInfo.ProductName + ' v.' + IntToStr(FVersionInfo.V1) {+ '.' +
                IntToStr(FVersionInfo.V2)};
  FDirPath := AppDataFolder + '\' + FVersionInfo.CompanyName + '\' +
               FProductID;
  if not (DirectoryExists(FDirPath)) then
    ForceDirectories(FDirPath);

  FFilePath := FDirPath + '\' + FOwnerForm.Name + '.ini';
  FXMLFilePath := FDirPath + '\' + FOwnerForm.Name + '.xml';
  FControlList := TList.Create;
  EnumerateControls(FOwnerForm);
  ReadStateFromFile;
end;

constructor TFormStateSaver.Create(OwnerForm: TTntForm; UseXML: Boolean;
                                  SaveCheckboxStates, SaveSpinEditValues,
                                  SaveComboBoxStates, SaveRadioGroupStates,
                                  SaveEditContents: Boolean);
begin
  FOwnerForm := OwnerForm;
  FUseXML := UseXML;
  FSaveCheckboxStates := SaveCheckboxStates;
  FSaveSpinEditValues := SaveSpinEditValues;
  FSaveComboBoxStates := SaveComboBoxStates;
  FSaveRadioGroupStates := SaveRadioGroupStates;
  FSaveEditContents := SaveEditContents;
//Build the application data path for storing the info
  FVersionInfo := TAppVersionInfo.Create;
  FProductID := FVersionInfo.ProductName + ' v.' + IntToStr(FVersionInfo.V1) {+ '.' +
                IntToStr(FVersionInfo.V2)};
  FDirPath := AppDataFolder + '\' + FVersionInfo.CompanyName + '\' +
               FProductID;
  if not (DirectoryExists(FDirPath)) then
    ForceDirectories(FDirPath);

  FFilePath := FDirPath + '\' + FOwnerForm.Name + '.ini';
  FXMLFilePath := FDirPath + '\' + FOwnerForm.Name + '.xml';
  FControlList := TList.Create;
  EnumerateControls(FOwnerForm);
  ReadStateFromFile;
end;

destructor TFormStateSaver.Destroy;
begin
  WriteStateToFile;
  FControlList.Free;
  inherited;
end;

procedure TFormStateSaver.WriteStateToFile;
begin
  if UseXML then
    WriteStateToXMLFile
  else
    WriteStateToINIFile;
end;

procedure TFormStateSaver.WriteStateToINIFile;
var
Ini: TIniFile;
i: integer;
j: integer;

begin
//Save the data to an INI file if the app has a directory to put it in.
  if not DirectoryExists(FDirPath) then
    Exit;
  Ini := TIniFile.Create(FFilePath);
  try
//Only write position and size if not maximized
    if not (FOwnerForm.WindowState = wsMaximized) then
      begin
        Ini.WriteInteger('Form', 'Left', FOwnerForm.Left);
        Ini.WriteInteger('Form', 'Top', FOwnerForm.Top);
        Ini.WriteInteger('Form', 'Width', FOwnerForm.Width);
        Ini.WriteInteger('Form', 'Height', FOwnerForm.Height);
      end;
    Ini.WriteBool('Form', 'Maximized', FOwnerForm.WindowState = wsMaximized);
//Store the monitor on which the form is located
    if Screen.MonitorCount > 0 then
      Ini.WriteInteger('Form', 'Monitor', FOwnerForm.Monitor.MonitorNum);

//We only care (at the moment, at any rate) about controls which have an Align
//setting which indicates that they're resizable.
    if FControlList.Count > 0 then
      for i := 0 to FControlList.Count - 1 do
        begin
//List view control needs to save its column widths. Don't save the last one;
//that will default to what remains
          if TControl(FControlList[i]) is TTntListView then
            with TControl(FControlList[i]) as TTntListView do
              begin
                if Columns.Count > 1 then
                  for j := 0 to Columns.Count - 2 do
                    Ini.WriteInteger('Controls', Name + '_Col_' + IntToStr(j), Columns[j].Width);
              end;
//Most panel sizes should be saved, but toolbars must be able to wrap
          if not (TControl(FControlList[i]) is TTntToolbar) then
            with TControl(FControlList[i]) do
              begin
                Case Align of
                  alLeft: Ini.WriteInteger('Controls', Name, Width);
                  alTop: Ini.WriteInteger('Controls', Name, Height);
                end;
              end;
//If the property is set, save the state of checkboxes on the form
          if (TControl(FControlList[i]) is TTntCheckbox) and SaveCheckboxStates then
            with TTntCheckBox(FControlList[i]) do
              Ini.WriteBool('Controls', Name, Checked);
//If the property is set, save the value of spinedit controls on the form
          if (TControl(FControlList[i]) is TSpinEdit) and SaveSpinEditValues then
            with TSpinEdit(FControlList[i]) do
              Ini.WriteInteger('Controls', Name, Value);
//If the property is set, save the itemindex of comboboxes on the form
          if (TControl(FControlList[i]) is TTntComboBox) and SaveComboBoxStates then
            with TTntComboBox(FControlList[i]) do
              Ini.WriteInteger('Controls', Name, ItemIndex);
//If the property is set, save the itemindex of radiogroups on the form
          if (TControl(FControlList[i]) is TTntRadioGroup) and SaveRadioGroupStates then
            with TTntRadioGroup(FControlList[i]) do
              Ini.WriteInteger('Controls', Name, ItemIndex);
//If the property is set, save the contents of edit controls on the form
          if (TControl(FControlList[i]) is TTntEdit) and SaveEditContents then
            with TTntEdit(FControlList[i]) do
              Ini.WriteString('Controls', Name, Text);
        end;
  finally
    Ini.Free;
  end;
end;

procedure TFormStateSaver.WriteStateToXMLFile;
var
DomImpl: TDomImplementation;
DomDoc: TDomDocument;
DomToXMLParser: TDomToXMLParser;
RootNode: TDomElement;
FormNode: TDomElement;
ControlsNode: TDomElement;
ChildNode: TDomElement;
i: integer;
j: integer;
Stream: TFileStream;

  procedure AppendChildNode(NodeName: WideString; NodeValue: WideString; TheParent: TDomElement);
  var
  TextNode: TDomText;
  begin
    ChildNode := TDomElement.Create(DomDoc, NodeName);
    TextNode := TDomText.Create(DomDoc);
    TextNode.NodeValue := NodeValue;
    ChildNode.AppendChild(TextNode);
    TheParent.AppendChild(ChildNode);
  end;

begin
//Save the data to an XML file if the app has a directory to put it in.
  if not DirectoryExists(FDirPath) then
    Exit;

  DomImpl := TDomImplementation.Create(nil);
  try
    DomDoc := TDomDocument.Create(DomImpl);
    DomToXMLParser := TDomToXMLParser.Create(nil);
    try
      DomToXMLParser.DomImpl := DomImpl;
      RootNode := TDomElement.Create(DomDoc, 'formState');
      DomDoc.AppendChild(RootNode);
      FormNode := TDomElement.Create(DomDoc, 'form');
      RootNode.AppendChild(FormNode);
      ControlsNode := TDomElement.Create(DomDoc, 'controls');
      RootNode.AppendChild(ControlsNode);
//Add the form data
//Only write position and size if not maximized
      if not (FOwnerForm.WindowState = wsMaximized) then
        begin
          AppendChildNode('left', IntToStr(FOwnerForm.Left), FormNode);
          AppendChildNode('top', IntToStr(FOwnerForm.Top), FormNode);
          AppendChildNode('width', IntToStr(FOwnerForm.Width), FormNode);
          AppendChildNode('height', IntToStr(FOwnerForm.Height), FormNode);
        end;
      AppendChildNode('maximized', BoolToStr(FOwnerForm.WindowState = wsMaximized), FormNode);
//Store the monitor on which the form is located
    if Screen.MonitorCount > 0 then
      AppendChildNode('monitor', IntToStr(FOwnerForm.Monitor.MonitorNum), FormNode);

//Now add data about individual controls.
//We only care (at the moment, at any rate) about controls which have an Align
//setting which indicates that they're resizable.
      if FControlList.Count > 0 then
        for i := 0 to FControlList.Count - 1 do
          begin
//List view control needs to save its column widths. Don't save the last one;
//that will default to what remains
            if TControl(FControlList[i]) is TTntListView then
              with TControl(FControlList[i]) as TTntListView do
                begin
                  if Columns.Count > 1 then
                    for j := 0 to Columns.Count - 2 do
                      AppendChildNode(Name + '_Col_' + IntToStr(j), IntToStr(Columns[j].Width), ControlsNode);
                end;
            if not (TControl(FControlList[i]) is TTntToolbar) then
              with TControl(FControlList[i]) do
                begin
                  Case Align of
                    alLeft: AppendChildNode(Name, IntToStr(Width), ControlsNode);
                    alTop: AppendChildNode(Name, IntToStr(Height), ControlsNode);
                  end;
                end;
//If the property is set, save the state of checkboxes on the form
            if (TControl(FControlList[i]) is TTntCheckbox) and SaveCheckboxStates then
              with TTntCheckBox(FControlList[i]) do
                AppendChildNode(Name, BoolToStr(Checked), ControlsNode);
//If the property is set, save the values of spinedits on the form
            if (TControl(FControlList[i]) is TSpinEdit) and SaveSpinEditValues then
              with TSpinEdit(FControlList[i]) do
                AppendChildNode(Name, IntToStr(Value), ControlsNode);
//If the property is set, save the itemindex of comboboxes on the form
            if (TControl(FControlList[i]) is TTntComboBox) and SaveComboBoxStates then
              with TTntComboBox(FControlList[i]) do
                AppendChildNode(Name, IntToStr(ItemIndex), ControlsNode);

//If the property is set, save the itemindex of radiogroups on the form
            if (TControl(FControlList[i]) is TTntRadioGroup) and SaveRadioGroupStates then
              with TTntRadioGroup(FControlList[i]) do
                AppendChildNode(Name, IntToStr(ItemIndex), ControlsNode);
//If the property is set, save the contents of edit controls on the form
            if (TControl(FControlList[i]) is TTntEdit) and SaveEditContents then
              with TTntEdit(FControlList[i]) do
                AppendChildNode(Name, Text, ControlsNode);

          end;
//Now save the file to disk
      Stream := TFileStream.Create(FXMLFilePath, fmCreate);
      try
        DomToXMLParser.WriteToStream(DomDoc, 'UTF-8', Stream);
      finally
        Stream.Free;
      end;
    finally
      DomToXMLParser.Free;
    end;
  finally
    DomImpl.Free;
  end;
end;

procedure TFormStateSaver.ReadStateFromFile;
begin
  if UseXML then
    ReadStateFromXMLFile
  else
    ReadStateFromINIFile;
end;

procedure TFormStateSaver.ReadStateFromINIFile;
var
Ini: TIniFile;
i: integer;
j: integer;
Mon: integer;

begin
//If there's an INI file, read it and set the appropriate properties of the
//form and its components.
  if FileExists(FFilePath) then
    begin
      Ini := TIniFile.Create(FFilePath);
      try
//First set Window size and position, to place it on the correct monitor
//and provide the correct bounds for a splash screen to display if it's
//using the form BoundsRect.
        FOwnerForm.Left := Ini.ReadInteger('Form', 'Left', FOwnerForm.Left);
        FOwnerForm.Top := Ini.ReadInteger('Form', 'Top', FOwnerForm.Top);
        FOwnerForm.Width := Ini.ReadInteger('Form', 'Width', FOwnerForm.Width);
        FOwnerForm.Height := Ini.ReadInteger('Form', 'Height', FOwnerForm.Height);
        FOwnerForm.Update;
//Now handle the situation where the bounds are outside the display available
//(which might happen if a second monitor was removed from the system.)
        if FOwnerForm.Left > (Screen.DesktopWidth - 50) then
          FOwnerForm.Left := 0;
        if FOwnerForm.Top > (Screen.DesktopHeight - 50) then
          FOwnerForm.Top := 0;

        if Ini.ReadBool('Form', 'Maximized', False) = True then
          begin
//Check which monitor it was on, if multi-monitor
            if Screen.MonitorCount > 0 then
              Mon := Ini.ReadInteger('Form', 'Monitor', 0);
            if (Mon > 0) and (Mon < Screen.MonitorCount) then
              FOwnerForm.MakeFullyVisible(Screen.Monitors[Mon]);
//Now maximize the window
            FOwnerForm.WindowState := wsMaximized;
            FOwnerForm.Update;
          end;

        if FControlList.Count > 0 then
          for i := 0 to FControlList.Count - 1 do
            begin
              if not (TControl(FControlList[i]) is TTntToolbar) then
                with TControl(FControlList[i]) do
                  begin
  //We only care (at the moment, at any rate) about controls which have an Align
  //setting which indicates that they're resizable.
                    Case Align of
                      alLeft: Width := Ini.ReadInteger('Controls', Name, Width);
                      alTop: Height := Ini.ReadInteger('Controls', Name, Height);
                    end;
                  end;
//List view control needs to reload its column widths. Don't bother with the last one;
//that will default to what remains
              if TControl(FControlList[i]) is TTntListView then
                with TControl(FControlList[i]) as TTntListView do
                  begin
                    if Columns.Count > 1 then
                      for j := 0 to Columns.Count - 2 do
                        Columns[j].Width := Ini.ReadInteger('Controls', Name + '_Col_' + IntToStr(j), Columns[j].Width);
//Trigger a resize event -- usually needed to shuffle the column widths to avoid scrollbars
                    if Assigned(onResize) then
                      begin
                        onResize(nil);
                        Application.ProcessMessages;
                      end;
                  end;
//If the property is set, read the state of checkboxes on the form
              if (TControl(FControlList[i]) is TTntCheckbox) and SaveCheckboxStates then
                with TTntCheckBox(FControlList[i]) do
                  Checked := Ini.ReadBool('Controls', Name, Checked);
//If the property is set, read the values of spinedits on the form
              if (TControl(FControlList[i]) is TSpinEdit) and SaveSpinEditValues then
                with TSpinEdit(FControlList[i]) do
                  Value := Ini.ReadInteger('Controls', Name, Value);
//If the property is set, read the itemindex of combo boxes on the form
              if (TControl(FControlList[i]) is TTntComboBox) and SaveComboBoxStates then
                with TTntComboBox(FControlList[i]) do
                  ItemIndex := Ini.ReadInteger('Controls', Name, ItemIndex);
//If the property is set, read the itemindex of radiogroups on the form
              if (TControl(FControlList[i]) is TTntRadioGroup) and SaveRadioGroupStates then
                with TTntRadioGroup(FControlList[i]) do
                  ItemIndex := Ini.ReadInteger('Controls', Name, ItemIndex);
//If the property is set, read the text of edit controls on the form
              if (TControl(FControlList[i]) is TTntEdit) and SaveEditContents then
                with TTntEdit(FControlList[i]) do
                  Text := Ini.ReadString('Controls', Name, Text);
            end;
      finally
        Ini.Free;
      end;
    end;
end;

procedure TFormStateSaver.ReadStateFromXMLFile;
var
i: integer;
j: integer;
Path: TFileName;
DomImpl: TDomImplementation;
DomDoc: TDomDocument;
XMLToDomParser: TXMLToDomParser;
El: TDomElement;
boolTemp: Boolean;
wsTemp: WideString;
wsBool: WideString;
wsInt: WideString;
Mon: integer;

begin
  if not FileExists(FXMLFilePath) then
    Exit;
  Path := TFileName(FXMLFilePath);

  DomImpl := TDomImplementation.Create(nil);
  try
    XMLToDomParser := TXMLToDomParser.Create(nil);
    try
      XMLToDomParser.DOMImpl := DomImpl;
      DomDoc := XMLToDomParser.FileToDom(Path);
    finally
      XMLToDomParser.Free;
    end;
//Now read the data from the dom. First get the form element.
    if DomDoc.GetElementsByTagName('form').Length > 0 then
      begin
        El := TDomElement(DomDoc.GetElementsByTagName('form').Item(0));
        with FOwnerForm do
          begin
            if El.GetElementsByTagName('left').Length > 0 then
              Left := StrToIntDef(El.GetElementsByTagName('left').Item(0).textContent, Left);
            if El.GetElementsByTagName('top').Length > 0 then
              Top := StrToIntDef(El.GetElementsByTagName('top').Item(0).textContent, Top);
            if El.GetElementsByTagName('width').Length > 0 then
              Width := StrToIntDef(El.GetElementsByTagName('width').Item(0).textContent, Width);
            if El.GetElementsByTagName('height').Length > 0 then
              Height := StrToIntDef(El.GetElementsByTagName('height').Item(0).textContent, Height);
            if El.GetElementsByTagName('maximized').Length > 0 then
              begin
                BoolTemp := StrToBoolDef(El.GetElementsByTagName('maximized').Item(0).textContent, False);
                if BoolTemp then
                  begin
//Check which monitor it was on, if multi-monitor
                    if Screen.MonitorCount > 0 then
                      begin
                        Mon := 0;
                        if El.GetElementsByTagName('monitor').Length > 0 then
                          Mon := StrToIntDef(El.GetElementsByTagName('monitor').Item(0).textContent, 0);
                      end;
                    if (Mon > 0) and (Mon < Screen.MonitorCount) then
                      FOwnerForm.MakeFullyVisible(Screen.Monitors[Mon]);
//Now maximize the window
                    FOwnerForm.WindowState := wsMaximized;
                    FOwnerForm.Update;
                  end;
              end;
          end;
      end;
    if FControlList.Count > 0 then
        for i := 0 to FControlList.Count - 1 do
          begin
            if not (TControl(FControlList[i]) is TTntToolbar) then
              with TControl(FControlList[i]) do
                begin
    //We only care (at the moment, at any rate) about controls which have an Align
    //setting which indicates that they're resizable.
                  Case Align of
                    alLeft:
                      begin
                        if DomDoc.GetElementsByTagName(Name).Length > 0 then
                          Width := StrToIntDef(DomDoc.GetElementsByTagName(Name).Item(0).textContent, Width);
                      end;

                    alTop:
                      begin
                        if DomDoc.GetElementsByTagName(Name).Length > 0 then
                          Height := StrToIntDef(DomDoc.GetElementsByTagName(Name).Item(0).textContent, Height);
                      end;
                  end;
                end;
//List view control needs to reload its column widths. Don't bother with the last one;
//that will default to what remains
              if TControl(FControlList[i]) is TTntListView then
                with TControl(FControlList[i]) as TTntListView do
                  begin
                    if Columns.Count > 1 then
                      for j := 0 to Columns.Count - 2 do
                        begin
                          wsTemp := Name + '_Col_' + IntToStr(j);
                          if DomDoc.GetElementsByTagName(wsTemp).Length > 0 then
                            begin
                              Columns[j].Width := StrToIntDef(DomDoc.GetElementsByTagName(wsTemp).Item(0).textContent, Columns[j].Width);
//Trigger a resize event -- usually needed to shuffle the column widths to avoid scrollbars
                              if Assigned(onResize) then
                                begin
                                  onResize(nil);
                                  Application.ProcessMessages;
                                end;
                            end;
                        end;
                  end;
//If the property is set, read the state of checkboxes on the form
              if (TControl(FControlList[i]) is TTntCheckbox) and SaveCheckboxStates then
                with TTntCheckBox(FControlList[i]) do
                  if DomDoc.GetElementsByTagName(Name).Length > 0 then
                    begin
                      wsBool := DomDoc.GetElementsByTagName(Name).Item(0).textContent;
                      Checked := StrToBoolDef(string(wsBool), Checked);
                    end;
//If the property is set, read the values of spinedits on the form
              if (TControl(FControlList[i]) is TSpinEdit) and SaveSpinEditValues then
                with TSpinEdit(FControlList[i]) do
                  if DomDoc.GetElementsByTagName(Name).Length > 0 then
                    begin
                      wsInt := DomDoc.GetElementsByTagName(Name).Item(0).textContent;
                      Value := StrToIntDef(string(wsInt), Value);
                    end;
//If the property is set, read the itemindex of combo boxes on the form
              if (TControl(FControlList[i]) is TTntComboBox) and SaveComboBoxStates then
                with TTntComboBox(FControlList[i]) do
                  if DomDoc.GetElementsByTagName(Name).Length > 0 then
                    begin
                      wsInt := DomDoc.GetElementsByTagName(Name).Item(0).textContent;
                      ItemIndex := StrToIntDef(string(wsInt), ItemIndex);
                    end;
//If the property is set, read the itemindex of radiogroups on the form
              if (TControl(FControlList[i]) is TTntRadioGroup) and SaveRadioGroupStates then
                with TTntRadioGroup(FControlList[i]) do
                  if DomDoc.GetElementsByTagName(Name).Length > 0 then
                    begin
                      wsInt := DomDoc.GetElementsByTagName(Name).Item(0).textContent;
                      ItemIndex := StrToIntDef(string(wsInt), ItemIndex);
                    end;
//If the property is set, read the text of edit controls on the form
              if (TControl(FControlList[i]) is TTntEdit) and SaveEditContents then
                with TTntEdit(FControlList[i]) do
                  if DomDoc.GetElementsByTagName(Name).Length > 0 then
                    Text := DomDoc.GetElementsByTagName(Name).Item(0).textContent;
          end;
  finally
    DomImpl.Free;
  end;
end;

procedure TFormStateSaver.EnumerateControls(Parent: TComponent);
var
i: integer;

begin
//Get a list of components owned by the component (recursive).
  if Parent.ComponentCount > 0 then
    for i := 0 to Parent.ComponentCount-1 do
      begin
        if Parent.Components[i] is TControl then
          FControlList.Add(Parent.Components[i]);
        EnumerateControls(Parent.Components[i]);
      end;
end;



end.
