//********************************************************************************************************************************
//*                                                                                                                              *
//*     Filesystem Dialogs - BASS Sampler unit  3delite 2006                                                                    *
//*     See FilesystemDialogs.txt for details                                                                                    *
//*                                                                                                                              *
//* Checks if there is an ID3v2 "SESC" frame in the (probably) audio file (.mp3 .mp2 .flac) if not, samples the file with BASS   *
//* (if available) then writes this sample back (creates a new) "SESC" frame and ID3v2 if needed.                                *
//*                                                                                                                              *
//* Two licenses are available if you like this library:                                                                         *
//* Shareware License: 50 Euros                                                                                                  *
//* Commercial License: 500 Euros                                                                                                *
//*                                                                                                                              *
//*     http://www.3delite.hu/Object Pascal Developer Resources/filesystemdialogs.html                                           *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit BASSSampler;

interface

Uses SysUtils, Classes, Windows, ID3v2, BASS;

type
  TSampleCache = Array of Byte;

    procedure CheckBASSAvailability;
    function StreamLoadSESC(Tag: TID3v2Tag): TSampleCache;
    function ID3v2_SetSESC(Tag: TID3v2Tag; SESC: TSampleCache; Force: Boolean): Boolean;

var
  BASS_Available: Boolean = False;
  BASS_LibHandle: THandle;
  t_BASS_MusicLoad: function(mem: BOOL; f: Pointer; offset, length, flags, freq: DWORD): DWORD; stdcall;
  t_BASS_StreamCreateFile: function(mem: BOOL; f: Pointer; offset, length, flags: DWORD): DWORD; stdcall;
  t_BASS_ChannelGetLength: function(handle: DWORD): Int64; stdcall;
  t_BASS_ChannelGetInfo: function(handle: DWORD; var info: BASS_CHANNELINFO):BOOL; stdcall;
  t_BASS_ChannelIsActive: function(handle: DWORD): DWORD; stdcall;
  t_BASS_ChannelGetData: function(handle: DWORD; buffer: Pointer; length: DWORD): DWORD; stdcall;
  t_BASS_ChannelGetPosition: function(handle: DWORD): QWORD; stdcall;
  t_BASS_MusicFree: procedure(Channel: Cardinal); stdcall;
  t_BASS_StreamFree: procedure(Channel: Cardinal); stdcall;
  t_BASS_GetDevice: function: DWORD; stdcall;
  t_BASS_ChannelPlay: function(handle: DWORD; restart: BOOL): BOOL; stdcall;
  t_BASS_ChannelStop: function(handle: DWORD): BOOL; stdcall;
  t_BASS_ChannelPause: function(handle: DWORD): BOOL; stdcall;

implementation

//Uses FileDialog;

procedure CheckBASSAvailability;
var
  ActDevice: Integer;
begin

    BASS_LibHandle := GetModuleHandle(PChar('bass.dll'));

    if BASS_LibHandle <> 0 then begin
        t_BASS_MusicLoad := GetProcAddress(BASS_LibHandle, PChar('BASS_MusicLoad'));
        t_BASS_StreamCreateFile := GetProcAddress(BASS_LibHandle, PChar('BASS_StreamCreateFile'));
        t_BASS_ChannelGetLength := GetProcAddress(BASS_LibHandle, PChar('BASS_ChannelGetLength'));
        t_BASS_ChannelGetInfo := GetProcAddress(BASS_LibHandle, PChar('BASS_ChannelGetInfo'));
        t_BASS_ChannelIsActive := GetProcAddress(BASS_LibHandle, PChar('BASS_ChannelIsActive'));
        t_BASS_ChannelGetData := GetProcAddress(BASS_LibHandle, PChar('BASS_ChannelGetData'));
        t_BASS_ChannelGetPosition := GetProcAddress(BASS_LibHandle, PChar('BASS_ChannelGetPosition'));
        t_BASS_MusicFree := GetProcAddress(BASS_LibHandle, PChar('BASS_MusicFree'));
        t_BASS_StreamFree := GetProcAddress(BASS_LibHandle, PChar('BASS_StreamFree'));
        t_BASS_GetDevice := GetProcAddress(BASS_LibHandle, PChar('BASS_GetDevice'));
        t_BASS_ChannelPlay := GetProcAddress(BASS_LibHandle, PChar('BASS_ChannelPlay'));
        t_BASS_ChannelStop := GetProcAddress(BASS_LibHandle, PChar('BASS_ChannelStop'));
        t_BASS_ChannelPause := GetProcAddress(BASS_LibHandle, PChar('BASS_ChannelPause'));
    end;

    if (BASS_LibHandle <> 0)
    AND (Assigned(t_BASS_MusicLoad))
    AND (Assigned(t_BASS_StreamCreateFile))
    AND (Assigned(t_BASS_ChannelGetLength))
    AND (Assigned(t_BASS_ChannelGetInfo))
    AND (Assigned(t_BASS_ChannelIsActive))
    AND (Assigned(t_BASS_ChannelGetData))
    AND (Assigned(t_BASS_ChannelGetPosition))
    AND (Assigned(t_BASS_MusicFree))
    AND (Assigned(t_BASS_StreamFree))
    AND (Assigned(t_BASS_GetDevice))
    AND (Assigned(t_BASS_ChannelPlay))
    AND (Assigned(t_BASS_ChannelStop))
    AND (Assigned(t_BASS_ChannelPause))
        then BASS_Available := True
        else BASS_Available := False;

    if BASS_Available then begin
        ActDevice := t_BASS_GetDevice;
        BASS_Available := ActDevice <> -1;
    end;

end;

function StreamLoadSESC(Tag: TID3v2Tag): TSampleCache;
var
  SESCStream: TStream;
  i: integer;
  Frame: String;
begin
    SetLength(Result, 0);
    if NOT Tag.loaded
        then Exit;
    SESCStream := TMemoryStream.Create;
    try
        for i := 1 to Tag.Frames.Count do begin
            Frame := Tag.Frames.getFrameID(i);
            if Frame = 'SESC' then begin
                if Tag.Frames.isUnsynced(i)
                    then Tag.Frames.sync(i);
                if Tag.Frames.isCompressed(i)
                    then Tag.Frames.decompress(i);
                Tag.getFrameData(i, SESCStream);
                SetLength(Result, SESCStream.Size);
                SESCStream.Seek(0, soFromBeginning);
                {$RANGECHECKS OFF}
                SESCStream.ReadBuffer(Pointer(Result)^, SESCStream.Size);
                Break;
            end;
        end;
    finally
        SESCStream.Free;
    end;
end;

function ID3v2_SetSESC(Tag: TID3v2Tag; SESC: TSampleCache; Force: Boolean): Boolean;
var
  SESCStream: TStream;
  i, tempIntv: integer;
  Frame: string;
begin
    Result := False;
{$RANGECHECKS OFF}
    if Tag.loaded
    OR Force
    then begin
        SESCStream := TMemoryStream.Create;
        try
            for i := 1 to Tag.Frames.Count do begin
                Frame := Tag.Frames.getFrameID(i);
                if Frame = 'SESC' then begin
                    Tag.Frames.Delete(i);
                    Break;
                end;
                frame := '';
            end;

            Tag.Frames.add('SESC');
            for i := 0 to High(SESC) - 1
                do SESCStream.Write(SESC[i], SizeOf(SESC[i]));
            Tag.setFrameData(Tag.Frames.Count, SESCStream, 0, false);
            tempIntv := Tag.Frames.compress(Tag.Frames.Count, 3);
            if tempIntv = 0
                then Tag.frames.setSaveFlagBytes(Tag.Frames.Count, 2, Tag.frames.SaveFlagBytes[Tag.Frames.Count,2] or 8);
            if tempIntv = 0
                then Result := True;
        finally
            SESCStream.Free;
        end;
    end;
end;

end.
