//********************************************************************************************************************************
//*                                                                                                                              *
//*     TSampleDisplay2D  3delite 2006-2007                                                                                          *
//*     See FilesystemDialogs.txt for details                                                                                    *
//*                                                                                                                              *
//* There is a threaded and ordinary functions. Both expect an existing bitmap that has .Width and .Height properties!           *
//* If AntiAliasingLevel = 0 the image is erased with SampleBackground else if AntiAliasingLevel >= 1 the sample is drawn onto   *
//* the existing picture with optional antialiasing.                                                                             *
//* SESC loader checks if there is an ID3v2 SESC frame in the (probably) audio file (.mp3 .mp2 .flac, atc.) if not and thumbnail *
//* database is specified it is checked for a SESC for the filename. If still no sample cache, samples the file with BASS if it  *
//* is available and is initialised in the process space.                                                                        *
//* If SaveSESC = True then writes this sample cache back to the file (and creates a new ID3v2 if needed) or to thumbnail        *
//* database if specified.                                                                                                       *
//*                                                                                                                              *
//* Two licenses are available if you like this library:                                                                         *
//* Shareware License: 50 Euros                                                                                                  *
//* Commercial License: 500 Euros                                                                                                *
//*                                                                                                                              *
//*     http://www.3delite.hu/Object Pascal Developer Resources/tsampledisplay3d.html                                            *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit SampleDisplay2D;

interface

Uses  Classes, Graphics, Windows, ID3v2, GR32, BASSSampler {, FilesystemDialogsDefs} {, Helper3delite} ;

type
  TSampleDisplayEvent = (sdeSampling, sdeDrawing, sdeStretching);

type
  TProgressCallback = function(FileName: String; Progress: Double; var AbortSampling: Boolean; EventType: TSampleDisplayEvent; User: Pointer): Boolean of object; stdcall;
  TFinishedCallback = function(FileName: String; var Pic: TBitmap32; User: Pointer): Boolean of object; stdcall;

//* Options
type
  TSamplePictureOptions = record
    SampleLeft: TColor32;               //* Sample's color for Left channel (alpha supported)
    SampleRight: TColor32;              //* Sample's color for Right channel (alpha supported)
    SampleBackground: TColor32;         //* Sample's color for background (alpha supported) sample will be drawn onto the image with alpha so it is not erased!
    Amplitude: Double;                  //* Peak amplitude 1 = no change
    AntiAliasingLevel: Double;          //* 0 - 1 No antialiasing, no alpha, 1 < antialiasing: Oversampling = Dimensions * AntiAliasingLevel (20 = 238MB!)
    SaveSESC: Boolean;                  //* Save SESC inside ID3v2 for .mp3 .mp2 .mp1. .mpa .flac
    ProgressCallback: TProgressCallback;//* Progress callback
    CallbackInterval: Cardinal;         //* Interval in ms for the ProgressCallback (setting it too small will probably slow things down suggested = 250ms)
    FinishedCallback: TFinishedCallback;//* Finished callback if Return value is False the bitmap will not be updated
    UseThumbDatabase: Boolean;          //* Use Filesystem Dialogs type thumbnail database file if it has SESC for the file (speed)
    ThumbDatabaseFileName: String;      //* Specify a thumbnail database file (else default filename will be used in that folder)
    SaveThumbDatabase: Boolean;         //* Save the SESC in the thumbnail database file
    User: Pointer;                      //* User data that you receive in ProgressCallback()
  end;

    //* The non-threaded version, it can be called directly
    function CreatePictureFromAudioFile(FileName: String; OutPic: TBitmap32; SPOptions: TSamplePictureOptions): Boolean; overload;
    function CreatePictureFromAudioFile(FileName: String; OutPic: Graphics.TBitmap; SPOptions: TSamplePictureOptions): Boolean; overload;
    //function CreatePictureFromAudioFileThreaded(FileName: String; var OutPic: Graphics.TBitmap; SPOptions: TSamplePictureOptions): Boolean;
    function SampleDataCache(FileName: String): TSampleCache;
    function CreatePictureFromSESC(SampleCache: TSampleCache; var OutPic: TBitmap32; SPOptions: TSamplePictureOptions): Boolean;

//* The TBitmap32 thread class
type
  TSampleDisplay2DThread = class(TThread)
  private
    tFileName: String;
    tOptions: TSamplePictureOptions;
    tSamplePic: TBitmap32;
    tOutPic: TBitmap32;
    tProgress: Double;
    tProcess: TSampleDisplayEvent;
    procedure Update;
    procedure CallFinishedErrorCallback;
    procedure CallProgressCallback;
    procedure CallFinishedCallback;
    function CreatePictureFromAudioFile(FileName: String; var OutPic: TBitmap32): Boolean;
    function SampleDataCache(FileName: String): TSampleCache;
  public
    Constructor Create(FileName: String; OutPic: TBitmap32; SPOptions: TSamplePictureOptions);
    Destructor Destroy; override;
  protected
  procedure Execute; override;
  published
  
end;

//* The TBitmap thread class
type
  TSampleDisplay2DTBitmapThread = class(TSampleDisplay2DThread)
  private
    tOut: Graphics.TBitmap;
    tTemp: TBitmap32;
    procedure Update;
  public
    Constructor Create(FileName: String; OutPic: Graphics.TBitmap; SPOptions: TSamplePictureOptions);
    Destructor Destroy;
  protected
  procedure Execute; override;
end;

const
  FILE_DEBUG    = 'C:\TSampleDisplay2DThread.log';

    procedure Register;

implementation

Uses SysUtils, Forms, BASS, GR32_Resamplers, GR32_Layers, ThumbDatabase; 

procedure Register;
begin
    //RegisterComponents('3delite', [TSampleDisplay2DThread]);
end;

Constructor TSampleDisplay2DThread.Create(FileName: String; OutPic: TBitmap32; SPOptions: TSamplePictureOptions);
begin
    FreeOnTerminate := True;
    tFileName := FileName;
    tOptions := SPOptions;
    tOutPic := OutPic;
    tProgress := 0;
    Inherited Create(True);
    Priority := tpLower;
    tSamplePic := TBitmap32.Create;
end;

Destructor TSampleDisplay2DThread.Destroy;
begin
    try
        if Assigned(tSamplePic)
            then FreeAndNil(tSamplePic);
    except
        //*
    end;
    Inherited Destroy;
end;

procedure TSampleDisplay2DThread.Execute;
begin
    If (NOT Terminated) then begin
        tOutPic.Lock;
        tSamplePic.Assign(tOutPic);
        tOutPic.Unlock;
    end;
    If (NOT Terminated) then begin
        if CreatePictureFromAudioFile(tFileName, tSamplePic) then begin
            if NOT Terminated then begin
                if Assigned(tOptions.FinishedCallback) then begin
                    Synchronize(CallFinishedCallback);
                end;
                if NOT Terminated
                    then Synchronize(Update);
            end;
        end else Synchronize(CallFinishedErrorCallback);
    end;
end;

procedure TSampleDisplay2DThread.Update;
begin
    try
        tOutPic.Lock;
        tOutPic.Assign(tSamplePic);
        tOutPic.Unlock;
    except
        {$IFDEF _DEBUG}
        //Showmessage('Exception: TSampleDisplay2DThread.Update()');
        {$ENDIF}
    end;
end;

procedure TSampleDisplay2DThread.CallFinishedErrorCallback;
var
  Null: TBitmap32;
begin
    try
        If NOT Terminated then begin
            Null := nil;
            tOptions.FinishedCallback(tFileName, Null, tOptions.User);
            Terminate;
        end;
    except
        {$IFDEF _DEBUG}
        //Showmessage('Exception: TSampleDisplay2DThread.CallProgressCallback()');
        {$ENDIF}
    end;
end;

procedure TSampleDisplay2DThread.CallProgressCallback;
var
  TerminateThread: Boolean;
begin
    try
        TerminateThread := False;
        If NOT Terminated
            then tOptions.ProgressCallback(tFileName, tProgress, TerminateThread, tProcess, tOptions.User);
        if TerminateThread
            then Terminate;
    except
        {$IFDEF _DEBUG}
        //Showmessage('Exception: TSampleDisplay2DThread.CallProgressCallback()');
        {$ENDIF}
    end;
end;

procedure TSampleDisplay2DThread.CallFinishedCallback;
begin
    try
        If NOT Terminated then begin
            if NOT tOptions.FinishedCallback(tFileName, tSamplePic, tOptions.User)
                then Terminate;
        end;
    except
        {$IFDEF _DEBUG}
        //Showmessage('Exception: TSampleDisplay2DThread.CallProgressCallback()');
        {$ENDIF}
    end;
end;

function TSampleDisplay2DThread.SampleDataCache(FileName: String): TSampleCache;
var
  i, k: Integer;
  GotData: Cardinal;
  Offset: Extended;
  SDSize: Int64;
  lSample, rSample: Extended;
  Channel: DWORD;
  Buffer: array [0..512] of Single;
  SampleAll: Boolean;
  StreamLength: Int64;
  ChanInfo: BASS_CHANNELINFO;
  Cycles1, Cycles2: Cardinal;
  Channels: Cardinal;
  PrevPercentDone: Cardinal;
  CurrentPosition: Int64;
  AbortSampling: Boolean;
  Ticks, PrevTicks: Cardinal;
  SampleCount: Integer;
begin
    Channel := 0;
    PrevPercentDone := 0;
    tProgress := 0;
    SampleAll := False;
    AbortSampling := False;
    PrevTicks := 0;
    tProcess := sdeSampling;

    try
        SetLength(Result, 0);


        if Channel = 0 then begin
            Channel := t_BASS_MusicLoad(FALSE, PChar(FileName), 0, 0, BASS_MUSIC_DECODE OR BASS_SAMPLE_FLOAT OR BASS_MUSIC_STOPBACK OR BASS_MUSIC_CALCLEN , 0);   // BASS_SAMPLE_MONO
            if Channel = 0
                then Channel := t_BASS_StreamCreateFile(False, PChar(FileName), 0, 0, BASS_STREAM_DECODE OR BASS_SAMPLE_FLOAT OR BASS_MP3_SETPOS);
        end;

        if Channel = 0
            then Exit;

        StreamLength := t_BASS_ChannelGetLength(Channel);

        try
            if StreamLength > High(Integer)
                then StreamLength := High(Integer) - 1;
        except
            //*
        end;
        if StreamLength < 64000 then begin
            SDSize := StreamLength + 10;
            SetLength(Result, SDSize);
            SampleAll := True;
        end else begin
            Offset := (StreamLength / 256);
            //* Thanx to Eva Freimannova
            //SetLength(Streams.Index[Index].SampleCache, Trunc(Offset / 2));
            SampleCount := Trunc(Offset / 2);
            if odd(SampleCount)
                then Inc(SampleCount);
            SetLength(Result, SampleCount);
        end;

        t_BASS_ChannelGetInfo(Channel, ChanInfo);

        i := 0;
        while (t_BASS_ChannelIsActive(Channel) <> BASS_ACTIVE_STOPPED) do begin

            try
                ZeroMemory(@Buffer, SizeOf(Buffer));
                if NOT SampleAll
                    then GotData := t_BASS_ChannelGetData(Channel, @Buffer, BASS_DATA_FFT512 OR BASS_DATA_FFT_INDIVIDUAL)
                    else GotData := t_BASS_ChannelGetData(Channel, @Buffer, 256 * 4 OR BASS_DATA_FLOAT);
            except
                //*
            end;

            if GotData < 1
                then Break;

            if NOT SampleAll then begin

                lSample := Buffer[0];
                rSample := Buffer[ChanInfo.chans - 1];

                Cycles1 := (GotData div 16);
                Cycles2 := (GotData div 16) - 1;
                Channels := (ChanInfo.chans - 1);

                for k := 1 to Cycles1
                    do lSample := lSample + Buffer[k * 2];
                for k := 1 to Cycles2
                    do rSample := rSample + Buffer[k * 2 + Channels];

                lSample := lSample * 16;
                rSample := rSample * 16;

                if lSample > 255
                    then lSample := 255;
                if rSample > 255
                    then rSample := 255;

                Result[i] := Trunc(lSample);
                Inc(i);
                Result[i] := Trunc(rSample);

            end else begin

                Channels := (ChanInfo.chans - 1);

                for k := 0 to 128 do begin
                    if Buffer[k * 2] > 255
                        then Buffer[k * 2] := 255;
                    Result[i * k * 2] := Trunc(Buffer[k * 2] * 100);
                end;

                for k := 0 to 127 do begin
                    if Buffer[k * 2 + Channels] > 255
                        then Buffer[k * 2 + Channels] := 255;
                    Result[i * k * 2 + 1] := Trunc(Buffer[k * 2 + 1] * 100);
                end;

                for k := 1 to 127
                    do Result[i * k * 2] := Trunc((Result[i * k * 2 - 1] + Result[i * k * 2] + Result[i * k * 2 + 1]) / 3);

                for k := 1 to 126
                    do Result[i * k * 2 + 1] := Trunc((Result[i * k * 2] + Result[i * k * 2 + 1] + Result[i * k * 2 + 2]) / 3);

            end;

            Inc(i);
            if (NOT SampleAll)
            AND (i > High(Result))
                then Break;
            if SampleAll
            AND (i * k > High(Result))
                then Break;

            if (i mod 40) = 0 then begin
                //* Progress callback
                if Assigned(tOptions.ProgressCallback) then begin
                    try
                        Ticks := GetTickCount;
                        //* Update the progress
                        if (Ticks - PrevTicks) > tOptions.CallbackInterval then begin
                            CurrentPosition := t_BASS_ChannelGetPosition(Channel);
                            tProgress := 100 * (CurrentPosition / StreamLength);
                            Synchronize(CallProgressCallback);
                            PrevTicks := Ticks;
                        end;
                    except
                        //*
                    end;
                end;
            end;

            if AbortSampling
                then Break;

        end;

        if Assigned(tOptions.ProgressCallback) then begin
            tProgress := 100;
            Synchronize(CallProgressCallback);
        end;

        t_BASS_MusicFree(Channel);
        t_BASS_StreamFree(Channel);

        if NOT SampleAll
            then SetLength(Result, i)
            else SetLength(Result, i * k);

    except
        try
            t_BASS_MusicFree(Channel);
            t_BASS_StreamFree(Channel);
        except
            {$IFDEF _DEBUG}
            //MessageDlg('Exception: SampleDataCache(' + FileName + ')', mtWarning, [mbOk], 0);
            {$ENDIF}
        end;
    end;

end;

function TSampleDisplay2DThread.CreatePictureFromAudioFile(FileName: String; var OutPic: TBitmap32): Boolean;
var
  Tag: TID3v2Tag;
  SampleCache: TSampleCache;
  GfxOffset: Integer;
  i: Integer;
  tempx, tempy: Int64;
  Scale: Extended;
  AmplitudeScale: Extended;
  LeftAmplitude, RightAmplitude, LineY1, LineY2: Word;
  AudioResamplerType: TKernelResampler;
  AudioLanczResampler: TLanczosKernel;
  Fext: String;
  Layer: TBitmapLayer;
  AALEvel: Double;
  ID: String;
  SESCStream: TStream;
  ThumbDatabase: TThumbDatabase;
  AbortSampling, SESCSaved: Boolean;
  Ticks, PrevTicks: Cardinal;
  TagError: Integer;
begin

    Result := False;
    SESCSaved := False;
    AbortSampling := False;
    tProgress := 0;
    PrevTicks := 0;

    Fext := AnsiUpperCase(ExtractFileExt(FileName));

    Tag := TID3v2Tag.Create;

    try
        ZeroMemory(@Layer, SizeOf(TBitmapLayer));
        ZeroMemory(@ThumbDatabase, SizeOf(TThumbDatabase));

        Tag.loadFromFile(FileName);
        SampleCache := StreamLoadSESC(Tag);

        if (Length(SampleCache) = 0)
        AND tOptions.UseThumbDatabase
        then begin
            //* Check thumbnail database
            ThumbDatabase := TThumbDatabase.Create(nil);

            if FileExists(tOptions.ThumbDatabaseFileName)
                then Result := ThumbDatabase.Load(tOptions.ThumbDatabaseFileName)
                else Result := ThumbDatabase.Load(ExtractFilePath(FileName) + FILE_THUMBNAILCACHE);

            if Result then begin
                SESCStream := TMemoryStream.Create;
                try
                    if ThumbDatabase.GetThumb(FileName, SESCStream, tempx, tempy, ID) then begin
                        if ID = 'SESC' then begin
                            SESCStream.Seek(0, soFromBeginning);
                            SetLength(SampleCache, SESCStream.Size);
                            SESCStream.ReadBuffer(Pointer(SampleCache)^, SESCStream.Size);
                            SESCSaved := True;
                        end;
                    end;
                finally
                    FreeAndNil(SESCStream);
                end;
                Result := False;
            end;
        end;

        //* BASS?
        if BASS_Available then begin
            if Length(SampleCache) < 1 then begin
                SampleCache := SampleDataCache(FileName);
                if tOptions.SaveSESC then begin
                    if (Fext = '.MP3')
                    OR (Fext = '.MP2')
                    OR (Fext = '.MP1')
                    OR (Fext = '.FLAC')
                    OR (Fext = '.AAC')
                    then if ID3v2_SetSESC(Tag, SampleCache, True) then begin
                        TagError := Tag.saveToFile(FileName);
                        if TagError < 256
                            then SESCSaved := True;
                        {$IFDEF _DEBUG_ID3V2}
                        if TagError > 0
                            then MessageDlg('CreateAudioThumbnail() Tag save error: ' + IntToStr(TagError), mtWarning, [mbOk], 0)
                        {$ENDIF}
                    end;
                end;
            end;
        end;

        if Length(SampleCache) < 1
            then Exit;

        if tOptions.SaveThumbDatabase
        AND (Length(SampleCache) > 0)
        AND (NOT SESCSaved)
        then begin

            if NOT Assigned(ThumbDatabase)
                then ThumbDatabase := TThumbDatabase.Create(nil);

            //* Load the database
            if FileExists(tOptions.ThumbDatabaseFileName)
                then Result := ThumbDatabase.Load(tOptions.ThumbDatabaseFileName)
                else Result := ThumbDatabase.Load(ExtractFilePath(FileName) + FILE_THUMBNAILCACHE);

            //* Save to thumbnail database
            SESCStream := TMemoryStream.Create;
            try
                SESCStream.Seek(0, soFromBeginning);
                SESCStream.WriteBuffer(Pointer(SampleCache)^, Length(SampleCache));
                ThumbDatabase.AddThumb(FileName, SESCStream, -1, -1, 'SESC', True, False);
                //* Save the thumbnail database cache
                //* Saves the thumbnail database every time a new thumbnail is added
                //* Looks like doesn't have real impact on performance
                if tOptions.SaveThumbDatabase
                AND (ThumbDatabase.fThumbCount > 0)
                    { TODO : Append path if not specified }
                    then if FileExists(tOptions.ThumbDatabaseFileName)
                        then ThumbDatabase.Save(tOptions.ThumbDatabaseFileName)
                        else ThumbDatabase.Save(ExtractFilePath(FileName) + FILE_THUMBNAILCACHE);
            finally
                FreeAndNil(SESCStream);
            end;
        end;

        if tOptions.AntiAliasingLevel < 1
            then AALevel := 1
            else AALevel := tOptions.AntiAliasingLevel;

        Layer := TBitmapLayer.Create(nil);
        Layer.Bitmap.Width := Trunc(OutPic.Width * AALevel);
        Layer.Bitmap.Height := Trunc(OutPic.Height * AALevel);

        AudioLanczResampler := TLanczosKernel.Create;
        AudioResamplerType := TKernelResampler.Create(Layer.Bitmap);
        AudioResamplerType.Kernel := AudioLanczResampler;
        AudioResamplerType.Changed;
        Layer.Bitmap.Resampler := AudioResamplerType;

        tProcess := sdeDrawing;

        try
            GfxOffset := Layer.Bitmap.Height div 2;
            AmplitudeScale := Layer.Bitmap.Height / 400;
            if tOptions.Amplitude <> 0
                then AmplitudeScale := AmplitudeScale * tOptions.Amplitude;

            Layer.Bitmap.HorzLineS(0, GfxOffset, Layer.Bitmap.Width, clWhite32);

            Scale := (Length(SampleCache) / Layer.Bitmap.Width / 2);

            for i := 1 to Layer.Bitmap.Width - 2 do begin
                if Trunc(i * 2 * Scale + 1) > High(SampleCache)
                    then Break;

                LeftAmplitude := Round(SampleCache[Trunc(i * 2 * Scale)]);
                RightAmplitude := Round(SampleCache[Trunc(i * 2 * Scale + 1)]);

                LeftAmplitude := Round(LeftAmplitude * AmplitudeScale);
                RightAmplitude := Round(RightAmplitude * AmplitudeScale);

                LineY1 := GfxOffset - LeftAmplitude;
                LineY2 := GfxOffset + RightAmplitude;
                try
                    Layer.Bitmap.SetStipple([tOptions.SampleLeft, tOptions.SampleRight]);
                    Layer.Bitmap.StippleStep := 100 / ((LineY2 - LineY1) / 0.0104);
                    Layer.Bitmap.StippleCounter := 0;
                    Layer.Bitmap.VertLineTSP(i, LineY1, LineY2);
                except
                    //*
                end;

                if (i mod 40) = 0 then begin
                    //* Progress callback
                    if Assigned(tOptions.ProgressCallback) then begin
                        Ticks := GetTickCount;
                        //* Update the progress
                        if (Ticks - PrevTicks) > tOptions.CallbackInterval then begin
                            tProgress := 100 * (i / (Layer.Bitmap.Width - 2));
                            Synchronize(CallProgressCallback);
                            PrevTicks := Ticks;
                        end;
                    end;
                end;
                if AbortSampling
                    then Break;
            end;

            //* Middle line
            Layer.Bitmap.LineAS(0, GfxOffset, Layer.Bitmap.Width, GfxOffset, $55FFFFFF);

        except
            //*
        end;

        //* Raw
        //Layer.Bitmap.SaveToFile(ChangeFileExt('C:\Layer_' + ExtractFileName(FileName), '.bmp'));

        //* Progress 100%
        if Assigned(tOptions.ProgressCallback) then begin
            tProgress := 100;
            Synchronize(CallProgressCallback);
        end;

        //* What to do with original?
        if tOptions.AntiAliasingLevel = 0 then begin
            OutPic.Clear(tOptions.SampleBackground);
        end;

        //* Stretch-filter
        tProcess := sdeStretching;
        Layer.Bitmap.DrawMode := dmBlend;  //dmOpaque
        OutPic.DrawMode := dmBlend;
        Layer.Bitmap.DrawTo(OutPic, OutPic.BoundsRect, Layer.Bitmap.BoundsRect);

        //* Oversampled final image
        //OutPic.SaveToFile(ChangeFileExt('C:\LayerOversampled_' + ExtractFileName(FileName), '.bmp'));

        if i < 1
            then Result := False
            else Result := True;

    finally
        SetLength(SampleCache, 0);
        if Assigned(Layer)
            then FreeAndNil(Layer);
        if Assigned(ThumbDatabase)
            then FreeAndNil(ThumbDatabase);
        Tag.Free;
    end;

end;

Constructor TSampleDisplay2DTBitmapThread.Create(FileName: String; OutPic: Graphics.TBitmap; SPOptions: TSamplePictureOptions);
begin
    tOut := OutPic;
    Inherited Create(FileName, tOutPic, SPOptions);
    tTemp := TBitmap32.Create;
end;

Destructor TSampleDisplay2DTBitmapThread.Destroy;
begin
    try
        if Assigned(tTemp)
            then FreeAndNil(tTemp);
    except
        //*
    end;
    Inherited Destroy;
end;

procedure TSampleDisplay2DTBitmapThread.Execute;
begin
    If (NOT Terminated) then begin
        try
            tTemp.Assign(tOut);
            tOutPic := tTemp;
        except
            Terminate;
        end;
    end;
    If (NOT Terminated)
        then Inherited Execute;
    If (NOT Terminated)
        then Update;
end;

procedure TSampleDisplay2DTBitmapThread.Update;
var
  TempStream: TStream;
begin
    Inherited Update;
    try
        TempStream := TMemoryStream.Create;
        tOutPic.SaveToStream(TempStream);
        TempStream.Seek(0, soFromBeginning);
        tOut.LoadFromStream(TempStream);
        TempStream.Free;
    except
        {$IFDEF _DEBUG}
        //Showmessage('Exception: TSampleDisplay2DThread.Update()');
        {$ENDIF}
    end;
end;

function SampleDataCache(FileName: String): TSampleCache;
var
  i, k: Integer;
  GotData: Cardinal;
  Offset: Extended;
  SDSize: Int64;
  lSample, rSample: Extended;
  Channel: DWORD;
  Buffer: array [0..512] of Single;
  SampleAll: Boolean;
  StreamLength: Int64;
  ChanInfo: BASS_CHANNELINFO;
  Cycles1, Cycles2: Cardinal;
  Channels: Cardinal;
  PrevPercentDone: Cardinal;
  AbortSampling: Boolean;
  //Ticks, PrevTicks: Cardinal;
begin
    Channel := 0;
    PrevPercentDone := 0;
    SampleAll := False;
    AbortSampling := False;
    try
        SetLength(Result, 0);


        if Channel = 0 then begin
            Channel := t_BASS_MusicLoad(FALSE, PChar(FileName), 0, 0, BASS_MUSIC_DECODE OR BASS_SAMPLE_FLOAT OR BASS_MUSIC_STOPBACK OR BASS_MUSIC_CALCLEN , 0);   // BASS_SAMPLE_MONO
            if Channel = 0
                then Channel := t_BASS_StreamCreateFile(False, PChar(FileName), 0, 0, BASS_STREAM_DECODE OR BASS_SAMPLE_FLOAT OR BASS_MP3_SETPOS);
        end;

        if Channel = 0
            then Exit;

        StreamLength := t_BASS_ChannelGetLength(Channel);

        try
            if StreamLength > High(Integer)
                then StreamLength := High(Integer) - 1;
        except
            //*
        end;
        if StreamLength < 64000 then begin
            SDSize := StreamLength + 10;
            SetLength(Result, SDSize);
            SampleAll := True;
        end else begin
            Offset := (StreamLength / 256);
            SetLength(Result, Trunc(Offset / 2));
        end;

        t_BASS_ChannelGetInfo(Channel, ChanInfo);

        i := 0;
        while (t_BASS_ChannelIsActive(Channel) <> BASS_ACTIVE_STOPPED) do begin

            try
                ZeroMemory(@Buffer, SizeOf(Buffer));
                if NOT SampleAll
                    then GotData := t_BASS_ChannelGetData(Channel, @Buffer, BASS_DATA_FFT512 OR BASS_DATA_FFT_INDIVIDUAL)
                    else GotData := t_BASS_ChannelGetData(Channel, @Buffer, 256 * 4 OR BASS_DATA_FLOAT);
            except
                //*
            end;

            if GotData < 1
                then Break;

            if NOT SampleAll then begin

                lSample := Buffer[0];
                rSample := Buffer[ChanInfo.chans - 1];

                Cycles1 := (GotData div 16);
                Cycles2 := (GotData div 16) - 1;
                Channels := (ChanInfo.chans - 1);

                for k := 1 to Cycles1
                    do lSample := lSample + Buffer[k * 2];
                for k := 1 to Cycles2
                    do rSample := rSample + Buffer[k * 2 + Channels];

                lSample := lSample * 16;
                rSample := rSample * 16;

                if lSample > 255
                    then lSample := 255;
                if rSample > 255
                    then rSample := 255;

                Result[i] := Trunc(lSample);
                Inc(i);
                Result[i] := Trunc(rSample);

            end else begin

                Channels := (ChanInfo.chans - 1);

                for k := 0 to 128 do begin
                    if Buffer[k * 2] > 255
                        then Buffer[k * 2] := 255;
                    Result[i * k * 2] := Trunc(Buffer[k * 2] * 100);
                end;

                for k := 0 to 127 do begin
                    if Buffer[k * 2 + Channels] > 255
                        then Buffer[k * 2 + Channels] := 255;
                    Result[i * k * 2 + 1] := Trunc(Buffer[k * 2 + 1] * 100);
                end;

                for k := 1 to 127
                    do Result[i * k * 2] := Trunc((Result[i * k * 2 - 1] + Result[i * k * 2] + Result[i * k * 2 + 1]) / 3);

                for k := 1 to 126
                    do Result[i * k * 2 + 1] := Trunc((Result[i * k * 2] + Result[i * k * 2 + 1] + Result[i * k * 2 + 2]) / 3);

            end;

            Inc(i);
            if (NOT SampleAll)
            AND (i > High(Result))
                then Break;
            if SampleAll
            AND (i * k > High(Result))
                then Break;

            if (i mod 40) = 0 then begin
                {
                if Assigned(tOptions.ProgressCallback) then begin
                    try
                        Ticks := GetTickCount;
                        //* Update the thumbnail progress a second
                        if (Ticks - PrevTicks) > tOptions.CallbackInterval then begin
                            CurrentPosition := t_BASS_ChannelGetPosition(Channel);
                            tProgress := 100 * (CurrentPosition / StreamLength);
                            Synchronize(CallProgressCallback);
                            PrevTicks := Ticks;
                        end;
                    except
                        //*
                    end;
                end;
                }
                Application.ProcessMessages;
            end;

            if AbortSampling
                then Break;

        end;

        t_BASS_MusicFree(Channel);
        t_BASS_StreamFree(Channel);

        if NOT SampleAll
            then SetLength(Result, i)
            else SetLength(Result, i * k);

    except
        try
            t_BASS_MusicFree(Channel);
            t_BASS_StreamFree(Channel);
        except
            {$IFDEF _DEBUG}
            //MessageDlg('Exception: SampleDataCache(' + FileName + ')', mtWarning, [mbOk], 0);
            {$ENDIF}
        end;
    end;

end;

function CreatePictureFromAudioFile(FileName: String; OutPic: TBitmap32; SPOptions: TSamplePictureOptions): Boolean;
var
  Tag: TID3v2Tag;
  SampleCache: TSampleCache;
  GfxOffset: Integer;
  i: Integer;
  tempx, tempy: Int64;
  Scale: Extended;
  AmplitudeScale: Extended;
  LeftAmplitude, RightAmplitude, LineY1, LineY2: Word;
  AudioResamplerType: TKernelResampler;
  AudioLanczResampler: TLanczosKernel;
  Fext: String;
  Layer: TBitmapLayer;
  AALEvel: Double;
  ID: String;
  SESCStream: TStream;
  ThumbDatabase: TThumbDatabase;
  AbortSampling: Boolean;
  TagError: Integer;
  SESCSaved: Boolean;
begin

    Result := False;
    AbortSampling := False;
    SESCSaved := False;

    Fext := AnsiUpperCase(ExtractFileExt(FileName));

    Tag := TID3v2Tag.Create;

    try
        ZeroMemory(@Layer, SizeOf(TBitmapLayer));
        ZeroMemory(@ThumbDatabase, SizeOf(TThumbDatabase));

        Tag.loadFromFile(FileName);
        SampleCache := StreamLoadSESC(Tag);

        if (Length(SampleCache) = 0)
        AND SPOptions.UseThumbDatabase
        then begin
            //* Check thumbnail database
            ThumbDatabase := TThumbDatabase.Create(nil);

            if FileExists(SPOptions.ThumbDatabaseFileName)
                then Result := ThumbDatabase.Load(SPOptions.ThumbDatabaseFileName)
                else Result := ThumbDatabase.Load(ExtractFilePath(FileName) + FILE_THUMBNAILCACHE);

            if Result then begin
                SESCStream := TMemoryStream.Create;
                try
                    if ThumbDatabase.GetThumb(FileName, SESCStream, tempx, tempy, ID) then begin
                        if ID = 'SESC' then begin
                            SESCStream.Seek(0, soFromBeginning);
                            SetLength(SampleCache, SESCStream.Size);
                            SESCStream.ReadBuffer(Pointer(SampleCache)^, SESCStream.Size);
                            SESCSaved := True;
                        end;
                    end;
                finally
                    FreeAndNil(SESCStream);
                end;
                Result := False;
            end;
        end;

        //* BASS?
        if BASS_Available then begin
            if Length(SampleCache) < 1 then begin
                SampleCache := SampleDataCache(FileName);
                if SPOptions.SaveSESC then begin
                    if (Fext = '.MP3')
                    OR (Fext = '.MP2')
                    OR (Fext = '.MP1')
                    OR (Fext = '.FLAC')
                    OR (Fext = '.AAC')
                    then begin
                        if ID3v2_SetSESC(Tag, SampleCache, True) then begin
                            TagError := Tag.saveToFile(FileName);
                            if TagError < 256
                                then SESCSaved := True;
                            {$IFDEF _DEBUG_ID3V2}
                            if TagError > 0
                                then MessageDlg('CreateAudioThumbnail() Tag save error: ' + IntToStr(TagError), mtWarning, [mbOk], 0)
                            {$ENDIF}
                        end;
                    end;
                end;
            end;
        end;

        if Length(SampleCache) < 1
            then Exit;

        if SPOptions.SaveThumbDatabase
        AND (Length(SampleCache) > 0)
        AND (NOT SESCSaved)
        then begin

            if NOT Assigned(ThumbDatabase)
                then ThumbDatabase := TThumbDatabase.Create(nil);

            //* Load the database
            if FileExists(SPOptions.ThumbDatabaseFileName)
                then Result := ThumbDatabase.Load(SPOptions.ThumbDatabaseFileName)
                else Result := ThumbDatabase.Load(ExtractFilePath(FileName) + FILE_THUMBNAILCACHE);

            //* Save to thumbnail database
            SESCStream := TMemoryStream.Create;
            try
                SESCStream.Seek(0, soFromBeginning);
                SESCStream.WriteBuffer(Pointer(SampleCache)^, Length(SampleCache));
                ThumbDatabase.AddThumb(FileName, SESCStream, -1, -1, 'SESC', True, False);
                //* Save the thumbnail database cache
                //* Saves the thumbnail database every time a new thumbnail is added
                //* Looks like doesn't have real impact on performance
                if SPOptions.SaveThumbDatabase
                AND (ThumbDatabase.fThumbCount > 0)
                    { TODO : Append path if not specified }
                    then if FileExists(SPOptions.ThumbDatabaseFileName)
                        then ThumbDatabase.Save(SPOptions.ThumbDatabaseFileName)
                        else ThumbDatabase.Save(ExtractFilePath(FileName) + FILE_THUMBNAILCACHE);
            finally
                FreeAndNil(SESCStream);
            end;
        end;

        if SPOptions.AntiAliasingLevel < 1
            then AALevel := 1
            else AALevel := SPOptions.AntiAliasingLevel;

        Layer := TBitmapLayer.Create(nil);
        Layer.Bitmap.Width := Trunc(OutPic.Width * AALevel);
        Layer.Bitmap.Height := Trunc(OutPic.Height * AALevel);

        AudioLanczResampler := TLanczosKernel.Create;
        AudioResamplerType := TKernelResampler.Create(Layer.Bitmap);
        AudioResamplerType.Kernel := AudioLanczResampler;
        AudioResamplerType.Changed;
        Layer.Bitmap.Resampler := AudioResamplerType;

        try
            GfxOffset := Layer.Bitmap.Height div 2;
            AmplitudeScale := Layer.Bitmap.Height / 400;
            if SPOptions.Amplitude <> 0
                then AmplitudeScale := AmplitudeScale * SPOptions.Amplitude;

            Layer.Bitmap.HorzLineS(0, GfxOffset, Layer.Bitmap.Width, clWhite32);

            Scale := (Length(SampleCache) / Layer.Bitmap.Width / 2);

            for i := 1 to Layer.Bitmap.Width - 2 do begin
                if Trunc(i * 2 * Scale + 1) > High(SampleCache)
                    then Break;

                LeftAmplitude := Round(SampleCache[Trunc(i * 2 * Scale)]);
                RightAmplitude := Round(SampleCache[Trunc(i * 2 * Scale + 1)]);

                LeftAmplitude := Round(LeftAmplitude * AmplitudeScale);
                RightAmplitude := Round(RightAmplitude * AmplitudeScale);

                LineY1 := GfxOffset - LeftAmplitude;
                LineY2 := GfxOffset + RightAmplitude;
                try
                    Layer.Bitmap.SetStipple([SPOptions.SampleLeft, SPOptions.SampleRight]);
                    Layer.Bitmap.StippleStep := 100 / ((LineY2 - LineY1) / 0.0104);
                    Layer.Bitmap.StippleCounter := 0;
                    Layer.Bitmap.VertLineTSP(i, LineY1, LineY2);
                except
                    //*
                end;

                if (i mod 40) = 0 then begin
                    {
                    if Assigned(SPOptions.ProgressCallback) then begin
                        Ticks := GetTickCount;
                        //* Update the thumbnail progress a second
                        if (Ticks - PrevTicks) > tOptions.CallbackInterval then begin
                            tProgress := 100 * (i / (Layer.Bitmap.Width - 2));
                            Synchronize(CallProgressCallback);
                            PrevTicks := Ticks;
                        end;
                    end;
                    }
                    Application.ProcessMessages;
                end;
                if AbortSampling
                    then Break;
            end;

            Layer.Bitmap.LineAS(0, GfxOffset, Layer.Bitmap.Width, GfxOffset, $55FFFFFF);

        except
            //*
        end;

        //* Raw
        //Layer.Bitmap.SaveToFile(ChangeFileExt('C:\Layer_' + ExtractFileName(FileName), '.bmp'));

        //* Stretch-filter
        Layer.Bitmap.DrawMode := dmBlend;
        OutPic.DrawMode := dmBlend;
        Layer.Bitmap.DrawTo(OutPic, OutPic.BoundsRect, Layer.Bitmap.BoundsRect);

        //* Oversampled final image
        //OutPic.SaveToFile(ChangeFileExt('C:\LayerOversampled_' + ExtractFileName(FileName), '.bmp'));

        if i < 1
            then Result := False
            else Result := True;

    finally
        SetLength(SampleCache, 0);
        if Assigned(Layer)
            then FreeAndNil(Layer);
        if Assigned(ThumbDatabase)
            then FreeAndNil(ThumbDatabase);
        Tag.Free;
    end;

end;

function CreatePictureFromAudioFile(FileName: String; OutPic: Graphics.TBitmap; SPOptions: TSamplePictureOptions): Boolean; overload;
var
  TempBitmap: TBitmap32;
  TempStream: TStream;
begin
    try
        try
            ZeroMemory(@TempBitmap, SizeOf(TBitmap32));
            ZeroMemory(@TempStream, SizeOf(TStream));
            TempBitmap := TBitmap32.Create;
            TempBitmap.Assign(OutPic);
            Result := CreatePictureFromAudioFile(FileName, TempBitmap, SPOptions);
            if Result then begin
                TempStream := TMemoryStream.Create;
                TempBitmap.SaveToStream(TempStream);
                TempStream.Seek(0, soFromBeginning);
                OutPic.LoadFromStream(TempStream);
            end;
        except
            Result := False;
        end;
    finally
        if Assigned(TempBitmap)
            then FreeAndNil(TempBitmap);
        if Assigned(TempStream)
            then FreeAndNil(TempStream);
    end;
end;

{
function CreatePictureFromAudioFileThreaded(FileName: String; var OutPic: Graphics.TBitmap; SPOptions: TSamplePictureOptions): Boolean;
var
  TempBitmap: TBitmap32;
  SampleDisplay2D: TSampleDisplay2DThread;
  FCallback: TFinishedCallback;

    function FinishedCallback(FileName: String; var Pic: TBitmap32; User: Pointer): Boolean; stdcall;
    var
      TempStream: TStream;
    begin
        try
            ZeroMemory(@TempStream, SizeOf(TStream));
            TempStream := TMemoryStream.Create;
            Pic.SaveToStream(TempStream);
            TempStream.Seek(0, soFromBeginning);
            Graphics.TBitmap(User).LoadFromStream(TempStream);
            Result := False;
            if Assigned(TempStream)
                then FreeAndNil(TempStream);
            if Assigned(Pic)
                then FreeAndNil(Pic);
        except
            //*
        end;
    end;

begin
    try
        try
            ZeroMemory(@TempBitmap, SizeOf(TBitmap32));
            ZeroMemory(@SampleDisplay2D, SizeOf(TSampleDisplay2DThread));
            TempBitmap := TBitmap32.Create;
            TempBitmap.Assign(OutPic);
            if NOT Assigned(SPOptions.FinishedCallback) then begin
                FCallback := TFinishedCallback(FinishedCallback);
                SPOptions.FinishedCallback := FCallback;
            end;
            SPOptions.User := OutPic;
            SampleDisplay2D := TSampleDisplay2DThread.Create(FileName, TempBitmap, SPOptions);
            SampleDisplay2D.Resume;
            Result := True;
        except
            Result := False;
        end;
    finally
        //*
    end;
end;
}

function CreatePictureFromSESC(SampleCache: TSampleCache; var OutPic: TBitmap32; SPOptions: TSamplePictureOptions): Boolean;
var
  GfxOffset: Integer;
  i: Integer;
  //tempx, tempy: Int64;
  Scale: Extended;
  AmplitudeScale: Extended;
  LeftAmplitude, RightAmplitude, LineY1, LineY2: Word;
  AudioResamplerType: TKernelResampler;
  AudioLanczResampler: TLanczosKernel;
  //Layer: TBitmapLayer;
  SamplePic: TBitmap32;
  AALEvel: Double;
  //ID: String;
  AbortSampling, SESCSaved: Boolean;
  //Ticks, PrevTicks: Cardinal;
begin

    Result := False;
    SESCSaved := False;
    AbortSampling := False;
    //PrevTicks := 0;

    try
        //ZeroMemory(@Layer, SizeOf(TBitmapLayer));
        ZeroMemory(@SamplePic, SizeOf(TBitmap32));

        if Length(SampleCache) < 1
            then Exit;

        if SPOptions.AntiAliasingLevel < 1
            then AALevel := 1
            else AALevel := SPOptions.AntiAliasingLevel;

        SamplePic := TBitmap32.Create;
        SamplePic.Width := Trunc(OutPic.Width * AALevel);
        SamplePic.Height := Trunc(OutPic.Height * AALevel);
        SamplePic.Clear(SPOptions.SampleBackground);

        AudioLanczResampler := TLanczosKernel.Create;
        AudioResamplerType := TKernelResampler.Create(SamplePic);
        AudioResamplerType.Kernel := AudioLanczResampler;
        AudioResamplerType.Changed;
        SamplePic.Resampler := AudioResamplerType;

        {
        Layer := TBitmapLayer.Create(nil);
        Layer.Bitmap.Width := Trunc(OutPic.Width * AALevel);
        Layer.Bitmap.Height := Trunc(OutPic.Height * AALevel);

        AudioLanczResampler := TLanczosKernel.Create;
        AudioResamplerType := TKernelResampler.Create(Layer.Bitmap);
        AudioResamplerType.Kernel := AudioLanczResampler;
        AudioResamplerType.Changed;
        Layer.Bitmap.Resampler := AudioResamplerType;
        }

        try
            GfxOffset := SamplePic.Height div 2;
            AmplitudeScale := SamplePic.Height / 400;
            if SPOptions.Amplitude <> 0
                then AmplitudeScale := AmplitudeScale * SPOptions.Amplitude;

            SamplePic.HorzLineS(0, GfxOffset, SamplePic.Width, clWhite32);

            Scale := (Length(SampleCache) / SamplePic.Width / 2);

            for i := 1 to SamplePic.Width - 2 do begin
                if Trunc(i * 2 * Scale + 1) > High(SampleCache)
                    then Break;

                LeftAmplitude := Round(SampleCache[Trunc(i * 2 * Scale)]);
                RightAmplitude := Round(SampleCache[Trunc(i * 2 * Scale + 1)]);

                LeftAmplitude := Round(LeftAmplitude * AmplitudeScale);
                RightAmplitude := Round(RightAmplitude * AmplitudeScale);

                LineY1 := GfxOffset - LeftAmplitude;
                LineY2 := GfxOffset + RightAmplitude;
                try
                    SamplePic.SetStipple([SPOptions.SampleLeft, SPOptions.SampleRight]);
                    SamplePic.StippleStep := 100 / ((LineY2 - LineY1) / 0.0104);
                    SamplePic.StippleCounter := 0;
                    SamplePic.VertLineTSP(i, LineY1, LineY2);
                except
                    //*
                end;
                {
                if (i mod 40) = 0 then begin
                    //* Progress callback
                    if Assigned(tOptions.ProgressCallback) then begin
                        Ticks := GetTickCount;
                        //* Update the progress
                        if (Ticks - PrevTicks) > tOptions.CallbackInterval then begin
                            tProgress := 100 * (i / (Layer.Bitmap.Width - 2));
                            Synchronize(CallProgressCallback);
                            PrevTicks := Ticks;
                        end;
                    end;
                end;
                if AbortSampling
                    then Break;
                }
            end;

            //* Middle line
            SamplePic.LineAS(0, GfxOffset, SamplePic.Width, GfxOffset, $55FFFFFF);

        except
            //*
        end;

        //* Raw
        //Layer.Bitmap.SaveToFile(ChangeFileExt('C:\Layer_' + ExtractFileName(FileName), '.bmp'));

        //* Progress 100%
        {
        if Assigned(tOptions.ProgressCallback) then begin
            tProgress := 100;
            Synchronize(CallProgressCallback);
        end;
        }
        //* What to do with original?
        {
        if SPOptions.AntiAliasingLevel = 0 then begin
            OutPic.Clear(SPOptions.SampleBackground);
        end;
        }

        //* Stretch-filter
        SamplePic.DrawMode := dmBlend;
        OutPic.DrawMode := dmBlend;
        SamplePic.DrawTo(OutPic, OutPic.BoundsRect, SamplePic.BoundsRect);

        //* Oversampled final image
        //OutPic.SaveToFile(ChangeFileExt('C:\LayerOversampled_' + ExtractFileName(FileName), '.bmp'));

        if i < 1
            then Result := False
            else Result := True;

    finally
        if Assigned(SamplePic)
            then FreeAndNil(SamplePic);
    end;

end;

end.
